////////////////////////////////////////////////////////////////////////
// Class:       CheckTimeOffset
// Plugin Type: analyzer (art v2_11_03)
// File:        CheckTimeOffset_module.cc
//
// Generated at Sat Nov 10 16:52:23 2018 by Tingjun Yang using cetskelgen
// from cetlib version v3_03_01.
////////////////////////////////////////////////////////////////////////

#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Principal/Run.h"
#include "art/Framework/Principal/SubRun.h"
#include "canvas/Utilities/InputTag.h"
#include "canvas/Persistency/Common/FindManyP.h"
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"
#include "larcore/CoreUtils/ServiceUtil.h"
#include "lardata/DetectorInfoServices/DetectorPropertiesService.h"
#include "lardataobj/RecoBase/Track.h"
#include "lardataobj/RecoBase/Hit.h"

namespace t962 {
  class CheckTimeOffset;
}


class t962::CheckTimeOffset : public art::EDAnalyzer {
public:
  explicit CheckTimeOffset(fhicl::ParameterSet const & p);
  // The compiler-generated destructor is fine for non-base
  // classes without bare pointers or other resource use.

  // Plugins should not be copied or assigned.
  CheckTimeOffset(CheckTimeOffset const &) = delete;
  CheckTimeOffset(CheckTimeOffset &&) = delete;
  CheckTimeOffset & operator = (CheckTimeOffset const &) = delete;
  CheckTimeOffset & operator = (CheckTimeOffset &&) = delete;

  // Required functions.
  void analyze(art::Event const & e) override;

private:

  unsigned int fTrackID;
  art::InputTag fTrackModuleLabel;
  // Declare member data here.

};


t962::CheckTimeOffset::CheckTimeOffset(fhicl::ParameterSet const & p)
  :
  EDAnalyzer(p),
  fTrackID(p.get<unsigned int>("TrackID",0)),
  fTrackModuleLabel(p.get<art::InputTag>("TrackModuleLabel","pmtrack"))
{}

void t962::CheckTimeOffset::analyze(art::Event const & e)
{
  art::Handle< std::vector<recob::Track> > trackListHandle;
  std::vector<art::Ptr<recob::Track> > tracklist;
  if (e.getByLabel(fTrackModuleLabel,trackListHandle))
    art::fill_ptr_vector(tracklist, trackListHandle);

  art::FindManyP<recob::Hit> fmht(trackListHandle, e, fTrackModuleLabel);

  auto const detProp = art::ServiceHandle<detinfo::DetectorPropertiesService const>()->DataFor(e);

  if (fTrackID >= tracklist.size()){
    throw cet::exception("CheckTimeOffset_module.cc")<<"fTrackID = "<<fTrackID<<" is invalid since there are "<<tracklist.size()<<" tracks.";
  }
  if (!fmht.isValid()){
    throw cet::exception("CheckTimeOffset_module.cc")<<"fmht is invalid.";
  }
  else{

    auto& hits = fmht.at(fTrackID);
    std::vector<std::vector<double>> hittime(2);
    for (auto & hit : hits){
      hittime[hit->WireID().Plane].push_back(hit->PeakTime());
    }
    if (!hittime[0].size() || !hittime[1].size()){
      throw cet::exception("CheckTimeOffset_module.cc")<<"hit vector is empty.";
    }
    for (size_t i = 0; i < 2; ++i){
      std::sort(hittime[i].begin(), hittime[i].end());
      std::cout<<"Plane "<<i<<" number of hits "<<hittime[i].size()<<" median of peak time "<<hittime[i][hittime[i].size()/2]<<" x ticks offset "<<detProp.GetXTicksOffset(i,0,0)<<std::endl;
    }
  }


}

DEFINE_ART_MODULE(t962::CheckTimeOffset)
