////////////////////////////////////////////////////////////////////////
//
// BaseFilter class:
// Algoritm to produce a filtered event file having
// events with at least 1 cluster crawler cluster
//
// corey.adams@yale.edu
//
////////////////////////////////////////////////////////////////////////
#ifndef BASEFILTER_H
#define BASEFILTER_H


#include "TH1D.h"
#include <string>

/// Framework includes
#include "art/Framework/Core/EDFilter.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "fhiclcpp/ParameterSet.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art_root_io/TFileService.h"
#include "art_root_io/TFileDirectory.h"
#include "messagefacility/MessageLogger/MessageLogger.h"
#include "canvas/Persistency/Common/FindMany.h"

#include "lardataobj/RecoBase/Cluster.h"

namespace filt {

class BaseFilter : public art::EDFilter  {

public:

  explicit BaseFilter(fhicl::ParameterSet const& pset);
  virtual ~BaseFilter();


  bool filter(art::Event& evt);
  void reconfigure(fhicl::ParameterSet const& p);
  void beginJob();


private:

  std::string fCluster_label;
  unsigned int fMinHits;

protected:


}; // class BaseFilter

//-------------------------------------------------
BaseFilter::BaseFilter(fhicl::ParameterSet const & pset)
: EDFilter(pset)
{
  this->reconfigure(pset);
}

//-------------------------------------------------
BaseFilter::~BaseFilter()
{
}

//-------------------------------------------------
void BaseFilter::reconfigure(fhicl::ParameterSet const& p)
{
  fCluster_label = p.get< std::string >("ClusterLabel");
  fMinHits = p.get< unsigned int >("MinimumHits");
}

//-------------------------------------------------
void BaseFilter::beginJob()
{
}

//-------------------------------------------------
bool BaseFilter::filter(art::Event &evt)
{

  art::Handle< std::vector<recob::Cluster> > clusterHandle;
  evt.getByLabel(fCluster_label, clusterHandle);


  for (size_t i = 0; i < clusterHandle -> size(); i ++) {
    if ( clusterHandle->at(i).View() ==  geo::kV &&
         clusterHandle->at(i).NHits() > fMinHits) {
      return true;
    }
  }
  return false;

} // bool

DEFINE_ART_MODULE(BaseFilter)

} //namespace filt

#endif // MATCHFILTER_H
