////////////////////////////////////////////////////////////////////////
// Class:       SignalNoiseAna
// Plugin Type: analyzer (art v3_05_01)
// File:        SignalNoiseAna_module.cc
//
// Generated at Sat Sep 19 08:41:53 2020 by Wanwei Wu using cetskelgen
// from cetlib version v3_10_00.
////////////////////////////////////////////////////////////////////////

#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Principal/Run.h"
#include "art/Framework/Principal/SubRun.h"
#include "canvas/Utilities/InputTag.h"
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art_root_io/TFileService.h"
#include "art/Utilities/make_tool.h"
#include "cetlib_except/exception.h"
#include "cetlib/search_path.h"

#include "larcore/Geometry/Geometry.h"
#include "larcoreobj/SimpleTypesAndConstants/RawTypes.h"
#include "lardata/DetectorInfoServices/DetectorClocksService.h"
#include "lardata/DetectorInfoServices/DetectorPropertiesService.h"
#include "lardataobj/RawData/raw.h"
#include "lardataobj/RawData/RawDigit.h"
#include "lardataobj/RecoBase/Wire.h"
#include "lardataobj/Simulation/SimChannel.h"
#include "larevt/CalibrationDBI/Interface/ChannelStatusProvider.h"
#include "larevt/CalibrationDBI/Interface/ChannelStatusService.h"

#include "TEfficiency.h"
#include "TFile.h"
#include <TROOT.h>
#include "TTree.h"
#include "TH1D.h"
#include "TH2D.h"
#include "TCanvas.h"
#include "TProfile.h"
#include "TRandom.h"

using namespace std;

class SignalNoiseAna;


class SignalNoiseAna : public art::EDAnalyzer {
public:
  explicit SignalNoiseAna(fhicl::ParameterSet const& p);
  // The compiler-generated destructor is fine for non-base
  // classes without bare pointers or other resource use.

  // Plugins should not be copied or assigned.
  SignalNoiseAna(SignalNoiseAna const&) = delete;
  SignalNoiseAna(SignalNoiseAna&&) = delete;
  SignalNoiseAna& operator=(SignalNoiseAna const&) = delete;
  SignalNoiseAna& operator=(SignalNoiseAna&&) = delete;

  // Required functions.
  void analyze(art::Event const& e) override;

private:

  // User's function
  void beginJob() override;
  void endJob() override;
  void calculateFFT(TH1D* h1_waveform, TH1D* h1_fft);
  void getSignal(const sim::SimChannel& simchannel, std::vector<int> & signal_starttick, std::vector<int> & signal_endtick, std::vector<double> & signal_max, std::vector<int> & signal_max_tdctick);
  float getCharge(const sim::SimChannel& simchannel, int starttick, int endtick);
  void getTdcCharge(const sim::SimChannel& simchannel, int starttick, int endtick, std::vector<float> & tdcchargevec);
  bool isSignalInWindow(int maxtick, int starttick, int endtick);
  void reset();

  // fhicl parameters
  std::string fRawDigitLabel;
  art::InputTag fWireProducerLabel;
  art::InputTag fCNNROIWireProducerLabel;
  art::InputTag fSimulationProducerLabel;
  bool fDoADCThreshold; // true: add Adc Threshold
  bool fDoCNNROI; // true: add cnn roi
  bool fDoFrequencyTest; // true: add FFT Frequecnty statistical test
  bool fUseFullWaveform;
  int fWindowSize;
  int fStrideSize;
  bool fDropLastWindow;
  bool fSaveWaveform; // true: save waveform ADC
  bool fUseFloat; // true: float for ADC; false: short for ADC
  bool fUseNumElectrons; // true/flase: signal according to "number of electrons"/"energy deposits"
  double fSignalNumElectronsLowCut;
  double fSignalNumElectronsHighCut;
  bool fSaveSignal;
  bool fSaveNoise;
  bool fUsePureNoise;
  bool fEvaluateWaveform;
  std::string fFrequencyTemplateFile;
  bool fSaveTree;
  std::vector<float> fAdcThresholdCut;
  std::vector<float> fROIScoreCut;
  bool fCheckWaveform;
  bool fRebin; // rebin with factor of 2
  
  // user's parameters
  int fNPlanes;
  int fNTicks;
  float fFFTBinWidth;

  TProfile* pf_frequency_noise_template[3];
  TProfile* pf_frequency_signal_template[3];

  // tree
  TTree* fTree;
  int run;
  int subrun;
  int event;
  
  // waveform
  float channeladc[480][2048]; 

  // pure signal
  TH1D* h1_signal[3];

  TH2D* h2_signalADC_sametick[3];
  TH2D* h2_signalADC_max[3];
  TH1D* h1_tickdiff_max[3];

  TProfile* pf_frequency_signal[3];
  TProfile* pf_frequency_signal_default[3];
  
  // pure noise
  TH1D* h1_noiserms[3];
  TH1D* h1_noisemaxadc[3];
  TH1D* h1_noisecnnscore[3];
  TProfile* pf_frequency_noise_default[3];
  TProfile* pf_frequency_noise[3];

  float noisefrequency_noise[480][10][100]; // for save noise (pure): a*a 
  float noisefrequency_signal[480][10][100]; // for save noise (pure): b*b
  float noisefrequency_cross[480][10][100]; // for save noise (pure): a*b

  // evaluation: signal or noise
  TH1D* h1_signalmaxadc_all[3];
  TH1D* h1_signalmaxadc_select[3];
  TEfficiency* eff_maxadc[3];
  TH1D* h1_maxadc_signal[3];
  TH1D* h1_maxadc_noise[3];
  TH1D* h1_maxadc_noise_pure[3];
  
  TH1D* h1_esummaxadc_all[3];
  TH1D* h1_esummaxadc_select[3];
  TEfficiency* eff_esummaxadc[3];
  TH1D* h1_esummaxadc_signal[3];
  TH1D* h1_esummaxadc_noise[3];
  TH1D* h1_esummaxadc_noise_pure[3];
  
  TH1D* h1_signalcnnroi_all[3];
  TH1D* h1_signalcnnroi_select[3];
  TEfficiency* eff_cnnroi[3];
  TH1D* h1_roiscore_signal[3];
  TH1D* h1_roiscore_noise[3];
  TH1D* h1_roiscore_noise_pure[3];

  TH1D* h1_esumcnnroi_all[3];
  TH1D* h1_esumcnnroi_select[3];
  TEfficiency* eff_esumcnnroi[3];
  TH1D* h1_esumroiscore_signal[3];
  TH1D* h1_esumroiscore_noise[3];
  TH1D* h1_esumroiscore_noise_pure[3];

  float waveformadcsum[480][10];
  float waveformchargesum[480][10];
  float waveformsignal[480][10];
  float frequency_noise[480][10][100]; // noise value*value
  float frequency_signal[480][10][100]; // signal value*value
  float frequency_cross[480][10][100]; // noise value * signal value
  
  float waveformtdccharge[480][10][200]; // fCheckWaveform
  float waveformmaxadc[480][10];
  float waveformcnnscore[480][10];

  std::vector<int> simchannelvec; // sim channel channel number per event
  std::vector<std::vector<int>> signalstarttickvec; // signals per each sim channel per event
  std::vector<std::vector<int>> signalendtickvec; // signals per each sim channel per event
  std::vector<std::vector<int>> signalmaxtickvec; // signals per each sim channel per event
  std::vector<std::vector<double>> signalmaxvec; // signals per each sim channel per event

};


SignalNoiseAna::SignalNoiseAna(fhicl::ParameterSet const& p)
  : EDAnalyzer{p}  // ,
  // More initializers here.
{
  cet::search_path sp("FW_SEARCH_PATH");

  fRawDigitLabel             = p.get<std::string>("RawDigitLabel", "");
  fWireProducerLabel         = p.get<art::InputTag>("WireProducerLabel", "");
  fCNNROIWireProducerLabel   = p.get<art::InputTag>("CNNROIWireProducerLabel");
  fSimulationProducerLabel   = p.get<art::InputTag>("SimulationProducerLabel","largeant");
  fDoADCThreshold            = p.get<bool>("DoADCThreshold", false);
  fDoCNNROI                  = p.get<bool>("DoCNNROI", false);
  fDoFrequencyTest           = p.get<bool>("DoFrequencyTest", true);
  fUseFullWaveform           = p.get<bool>("UseFullWaveform", true);
  fWindowSize                = p.get<int>("ShortWaveformSize", 200);
  fStrideSize                = p.get<int>("StrideSize", 150);
  fDropLastWindow            = p.get<bool>("DropLastWindow", true);
  fSaveWaveform              = p.get<bool>("SaveWaveform", false);
  fUseFloat                  = p.get<bool>("UseFloat", true);
  fUseNumElectrons           = p.get<bool>("UseNumElectrons", true);
  fSignalNumElectronsLowCut  = p.get<double>("SignalNumElectronsLowCut");
  fSignalNumElectronsHighCut = p.get<double>("SignalNumElectronsHighCut");
  fSaveSignal                = p.get<bool>("SaveSignal", false);
  fSaveNoise                 = p.get<bool>("SaveNoise", false);
  fUsePureNoise              = p.get<bool>("UsePureNoise", false);
  fEvaluateWaveform          = p.get<bool>("EvaluateWaveform", false);
  fFrequencyTemplateFile     = sp.find_file(p.get<std::string>("FrequencyTemplateFile"));
  fSaveTree                  = p.get<bool>("SaveTree", false);
  fAdcThresholdCut           = p.get<std::vector<float>>("AdcThresholdCut");
  fROIScoreCut               = p.get<std::vector<float>>("ROIScoreCut");
  fCheckWaveform             = p.get<bool>("CheckWaveform", false);
  fRebin                     = p.get<bool>("Rebin", false);

  // use either raw waveform or recob::Wire waveform
  if (fRawDigitLabel.empty() && fWireProducerLabel.empty()){
    throw cet::exception("SignalNoiseAna") << "Both RawDigitLabel and WireProducerLabel are empty";
  }

  if ((!fRawDigitLabel.empty()) && (!fWireProducerLabel.empty())){
    throw cet::exception("SignalNoiseAna") << "Only one of RawDigitLabel and WireProducerLabel should be set";
  }

  fNPlanes = lar::providerFrom<geo::Geometry>()->Nplanes();
  if (fNPlanes > 3) {
    throw cet::exception("SignalNoiseAna") << "Number of planes is " << fNPlanes << "; please check";
  }

  auto const clockData = art::ServiceHandle<detinfo::DetectorClocksService>()->DataForJob();
  auto const detProp = art::ServiceHandle<detinfo::DetectorPropertiesService>()->DataForJob(clockData);
  fNTicks = detProp.NumberTimeSamples(); // number of clock ticks per event
  float fSampleRate = sampling_rate(clockData); // period of the TPC readout electronics clock [us]; 1 tick = 1 tdc = 198 ns.
  cout << "Numer of clock ticks per event: " << fNTicks << endl;
  cout << "Sampling rate: " << fSampleRate << endl;
  // width of frequency bin in kHz (signal window is 2048 here)
  fFFTBinWidth = 1.0/(fNTicks*fSampleRate*1.0e-6); // [kHz]
  cout << "fFFTBinWidth: " << fFFTBinWidth << endl;
}


void SignalNoiseAna::analyze(art::Event const& e)
{
  reset();

  run = e.run();
  subrun = e.subRun();
  event = e.id().event();

  auto const* geo = lar::providerFrom<geo::Geometry>();
  auto const clockData = art::ServiceHandle<detinfo::DetectorClocksService>()->DataFor(e);
  auto const detProp = art::ServiceHandle<detinfo::DetectorPropertiesService>()->DataFor(e, clockData);
  auto const& chStatus = art::ServiceHandle<lariov::ChannelStatusService>()->GetProvider();
  
  // get RawDigit
  art::Handle< std::vector<raw::RawDigit> > rawdigitListHandle;
  std::vector< art::Ptr<raw::RawDigit> > rawdigitlist;
  if (e.getByLabel(fRawDigitLabel, "", rawdigitListHandle)) {
    art::fill_ptr_vector(rawdigitlist, rawdigitListHandle);
  }

  // get wire
  art::Handle<std::vector<recob::Wire>> wireListHandle;
  std::vector<art::Ptr<recob::Wire>> wirelist;
  if (e.getByLabel(fWireProducerLabel, wireListHandle)) {
    art::fill_ptr_vector(wirelist, wireListHandle);
  }

  // get cnnroi wire
  art::Handle<std::vector<recob::Wire>> cnnroiwireListHandle;
  std::vector<art::Ptr<recob::Wire>> cnnroiwirelist;
  if (e.getByLabel(fCNNROIWireProducerLabel, cnnroiwireListHandle)) {
    art::fill_ptr_vector(cnnroiwirelist, cnnroiwireListHandle);
  }

  auto simChannelHandle = e.getValidHandle<std::vector<sim::SimChannel>>(fSimulationProducerLabel);

  if (fSaveSignal) {
    // ... Loop over simChannels
    for (auto const& simchannel : (*simChannelHandle)) {

      // .. get simChannel channel number
      const raw::ChannelID_t channel = simchannel.Channel();
      if (chStatus.IsBad(channel)) 
        continue;

      int view = geo->View(channel);

      std::vector<int> signal_starttick;
      std::vector<int> signal_endtick;
      std::vector<double> signal_max; // according to number of electrons or energy deposits
      std::vector<int> signal_max_tdctick;

      getSignal(simchannel, signal_starttick, signal_endtick, signal_max, signal_max_tdctick);

      if (signal_starttick.size() == 0 ||
          (signal_endtick.size() == 1 && signal_endtick.back() == -999))
        continue;
      
      // ---- signal ADC ----
      std::vector<float> adcvec(fNTicks);
      // recob wire
      if (wirelist.size()) {
        for (size_t w = 0; w < wirelist.size(); w++) {
          if (wirelist[w]->Channel() != channel) continue;

          std::vector<float> vecADC =
            wirelist[w]->Signal(); // a zero-padded full length vector filled with RoI signal.

          if (fUseFloat) {
            for (size_t j = 0; j < vecADC.size(); ++j) {
              adcvec[j] = vecADC[j];
            }
          }
          else {
            for (size_t j = 0; j < vecADC.size(); ++j) {
              adcvec[j] = short(vecADC[j]);
            }
          }
        } // loop over wirelist w
      } // if recob wire
      else if (rawdigitlist.size()) {
        for (size_t r = 0; r < rawdigitlist.size(); r++) {
          if (rawdigitlist[r]->Channel() != channel) continue;

          raw::RawDigit::ADCvector_t rawadc(fNTicks);
          raw::Uncompress(rawdigitlist[r]->ADCs(), rawadc, rawdigitlist[r]->Compression());

          float pedestal = rawdigitlist[r]->GetPedestal();
          for (size_t j=0; j < rawadc.size(); ++j) {
            adcvec[j] = rawadc[j] - pedestal;
          }
        } // loop over rawdigitlist r
      } // if rawdigit
      
      // "max" signal
      std::vector<float>::iterator temp_index = std::max_element(adcvec.begin(), adcvec.end());
      float max_adc = *temp_index;
      int max_adc_tick = std::distance(adcvec.begin(), temp_index);

      int max_signal_index = std::distance(signal_max.begin(), std::max_element(signal_max.begin(), signal_max.end()));
      
      h1_signal[view]->Fill(signal_max[max_signal_index]);
      h2_signalADC_sametick[view]->Fill(signal_max[max_signal_index], adcvec[signal_max_tdctick[max_signal_index]]);
      h2_signalADC_max[view]->Fill(signal_max[max_signal_index], max_adc);
      h1_tickdiff_max[view]->Fill(signal_max_tdctick[max_signal_index] - max_adc_tick);

      // signal FFT
      if (signal_max[max_signal_index] < fSignalNumElectronsLowCut) continue;
      
      if (fUseFullWaveform) {
        if (fSignalNumElectronsHighCut != -1 && signal_max[max_signal_index] > fSignalNumElectronsHighCut) continue;

        TH1D* h1_waveform = new TH1D(Form("waveform_%d", channel), Form("waveform_%d", channel), fNTicks, 0, fNTicks);
        for (int j = 0; j < fNTicks; ++j) {
          h1_waveform->SetBinContent(j+1, adcvec[j]);
        }
        TH1D* h1_fft = new TH1D(Form("fft_%d", channel), Form("fft_%d", channel), fNTicks, 0, fNTicks*fFFTBinWidth);

        // do FFT for single waveform
        calculateFFT(h1_waveform, h1_fft); 

        // save for each plane
        for (int j=0; j < h1_fft->GetNbinsX(); ++j) {
          pf_frequency_signal[view]->Fill((j + 0.5) * fFFTBinWidth, h1_fft->GetBinContent(j+1), 1);
          pf_frequency_signal_default[view]->Fill((j + 0.5) * fFFTBinWidth, h1_fft->GetBinContent(j+1), 1);
        }

        delete h1_waveform;
        delete h1_fft;
      } // fUseFullWaveform
      else {
        // loop over signals
        int tdc_start = -1;
        int tdc_end = -1;
        for (size_t s = 0; s < signal_starttick.size(); ++s) {

          if (signal_max[s] < fSignalNumElectronsLowCut) continue;
          if (fSignalNumElectronsHighCut != -1 && signal_max[s] > fSignalNumElectronsHighCut) continue;

          int start_s = s;
          int last_s = s;

          // check the following signals
          for (size_t s2 = s + 1; s2 < signal_starttick.size(); ++s2) {
            
            if (signal_max[s2] < fSignalNumElectronsLowCut) continue;
            if (fSignalNumElectronsHighCut != -1 && signal_max[s2] > fSignalNumElectronsHighCut) continue;

            if (signal_max_tdctick[s2] >= signal_max_tdctick[s] + fWindowSize) {
              s = s2 - 1;
              break;
            }
            else {
              last_s = s2;
              if (s2 == signal_starttick.size() - 1) s = s2; 
            }
          } // loop over signal s2
          
          // select a random range based on dt
          int dt = signal_max_tdctick[last_s] - signal_max_tdctick[start_s];
          
          tdc_start = signal_max_tdctick[start_s] - ((fWindowSize - dt - 0.1 * fWindowSize) / 2) * gRandom->Uniform(0,1);

          if (tdc_start < tdc_end) tdc_start = tdc_end + 1; // do not overlap with previous one
          if (tdc_start < 0) tdc_start = 0;

          tdc_end = tdc_start + fWindowSize - 1;
          if (tdc_end > fNTicks - 1) {
            tdc_end = fNTicks -1;
            tdc_start = tdc_end - fWindowSize + 1;
            // ideally, another check with previous one is needed here, but this should be very rare
          }

          // short waveform and FFT
          TH1D* h1_waveform = new TH1D(Form("waveform_%d", channel), Form("waveform_%d", channel), fWindowSize, 0, fWindowSize);
          for (int j = tdc_start; j <= tdc_end; ++j) {
            h1_waveform->SetBinContent(j-tdc_start+1, adcvec[j]);
          }
          TH1D* h1_fft = new TH1D(Form("fft_%d", channel), Form("fft_%d", channel), fWindowSize, 0, fWindowSize*fFFTBinWidth);
        
          // do FFT for each short waveform
          calculateFFT(h1_waveform, h1_fft); 

          // save for each plane
          if (!fRebin) {
            for (int j=0; j < h1_fft->GetNbinsX(); ++j) {
              pf_frequency_signal[view]->Fill((j + 0.5) * fFFTBinWidth, h1_fft->GetBinContent(j+1), 1);
              pf_frequency_signal_default[view]->Fill((j + 0.5) * fFFTBinWidth, h1_fft->GetBinContent(j+1), 1);
            }
          }
          else {
            for (int j=0; j < h1_fft->GetNbinsX()/2; ++j) {
              pf_frequency_signal[view]->Fill((j + 0.5) * fFFTBinWidth*2, h1_fft->GetBinContent(2*j+1) + h1_fft->GetBinContent(2*j+2), 1);
              pf_frequency_signal_default[view]->Fill((j + 0.5) * fFFTBinWidth*2, h1_fft->GetBinContent(j*2+1) + h1_fft->GetBinContent(2*j+2), 1);
            }
          }

          delete h1_waveform;
          delete h1_fft;

        } // loop over signal s
      } // use short waveform

    } // loop over simChannels
  } // fSaveSignal
  
  // ----------- noise -----------------
  if (fSaveNoise) {
    std::map<raw::ChannelID_t, bool> signalMap;
    for (auto const& channel : (*simChannelHandle)) {
      signalMap[channel.Channel()] = true;
    }

    // for AdcThrehold and/or FrequencyTest
    for (size_t i = 0; i < (rawdigitlist.empty() ? wirelist.size() : rawdigitlist.size()); ++i) {
      std::vector<float> adcvec(fNTicks);

      int channel = -1;

      if (rawdigitlist.size()) {
        channel = rawdigitlist[i]->Channel();
        if (!fUsePureNoise && signalMap[channel]) continue;
        if (chStatus.IsBad(channel)) continue;

        std::vector<short> rawadc(fNTicks); // vector to hold uncompressed adc values later
        raw::Uncompress(rawdigitlist[i]->ADCs(), rawadc, rawdigitlist[i]->GetPedestal(), rawdigitlist[i]->Compression());
        for (size_t j = 0; j < rawadc.size(); ++j) {
          adcvec[j] = rawadc[j] - rawdigitlist[i]->GetPedestal();
        }
      }
      else if (wirelist.size()) {
        channel = wirelist[i]->Channel();
        if (!fUsePureNoise && signalMap[channel]) continue;
        if (chStatus.IsBad(channel)) continue;

        const auto& signal = wirelist[i]->Signal();
        if (fUseFloat) {
          for (size_t j = 0; j < adcvec.size(); ++j) {
            adcvec[j] = signal[j];
          }
        }
        else {
          for (size_t j = 0; j < adcvec.size(); ++j) {
            adcvec[j] = short(signal[j]);
          }
        }
      }

      int view = geo->View(channel);

      // noise rms
      double temp_sum = 0.;
      for (size_t j = 0; j < adcvec.size(); j++) {
        temp_sum +=  adcvec[j] * adcvec[j];
      }
      h1_noiserms[view]->Fill(sqrt(temp_sum / adcvec.size()));

      // waveform and frequency
      if (fUseFullWaveform) {
        if (fDoADCThreshold) {
          float maxadc = 0;
          for (size_t j = 0; j < adcvec.size(); j++) {
            if (adcvec[j] > maxadc) maxadc = adcvec[j];
          }
          if (maxadc >= 20) maxadc = 20 - 1e-4; // overflow
          h1_noisemaxadc[view]->Fill(maxadc);
        } // fDoADCThreshold

        if (fDoFrequencyTest) {
          TH1D* h1_waveform = new TH1D(Form("waveform_%d", channel), Form("waveform_%d", channel), fNTicks, 0, fNTicks);
          for (int j = 0; j < fNTicks; ++j) {
            h1_waveform->SetBinContent(j+1, adcvec[j]);
          }
          TH1D* h1_fft = new TH1D(Form("fft_%d", channel), Form("fft_%d", channel), fNTicks, 0, fNTicks*fFFTBinWidth);

          // do FFT for single waveform
          calculateFFT(h1_waveform, h1_fft); 

          // save for each plane
          for (int j=0; j < h1_fft->GetNbinsX(); ++j) {
            pf_frequency_noise[view]->Fill((j + 0.5) * fFFTBinWidth, h1_fft->GetBinContent(j+1), 1);
            pf_frequency_noise_default[view]->Fill((j + 0.5) * fFFTBinWidth, h1_fft->GetBinContent(j+1), 1);
          }

          delete h1_waveform;
          delete h1_fft;
        } // fDoFrequencyTest
      } // if fUseFullWaveform
      else {
        int nstrides = int(fNTicks/fStrideSize+1);
        if (fDropLastWindow) nstrides = nstrides - 1;

        for (int k=0; k<nstrides; k++) {
          int start_tick = k*fStrideSize;
          int end_tick = start_tick + fWindowSize - 1; 

          if (fDoADCThreshold) {
            float maxadc = 0.;
            for (int j=start_tick; j<=end_tick; j++) {
              if (adcvec[j] > maxadc) maxadc = adcvec[j];
            }
            if (maxadc >= 20) maxadc = 20 - 1e-4; // overflow
            h1_noisemaxadc[view]->Fill(maxadc);
          } // fDoADCThreshold
          
          if (fDoFrequencyTest) {
            TH1D* h1_waveform = new TH1D(Form("waveform_%d", channel), Form("waveform_%d", channel), fWindowSize, 0, fWindowSize);
            for (int j=start_tick; j<=end_tick; ++j) {
              h1_waveform->SetBinContent(j-start_tick+1, adcvec[j]);
            }
            TH1D* h1_fft = new TH1D(Form("fft_%d", channel), Form("fft_%d", channel), fWindowSize, 0, fWindowSize*fFFTBinWidth);

            // do FFT for each short waveform
            calculateFFT(h1_waveform, h1_fft); 

            // save for each plane
            if (!fRebin) {
              for (int j=0; j < h1_fft->GetNbinsX(); ++j) {
                pf_frequency_noise[view]->Fill((j + 0.5) * fFFTBinWidth, h1_fft->GetBinContent(j+1), 1);
                pf_frequency_noise_default[view]->Fill((j + 0.5) * fFFTBinWidth, h1_fft->GetBinContent(j+1), 1);
              }
            }
            else {
              for (int j=0; j < h1_fft->GetNbinsX()/2; ++j) {
                pf_frequency_noise[view]->Fill((j + 0.5) * fFFTBinWidth*2, h1_fft->GetBinContent(2*j+1) + h1_fft->GetBinContent(2*j+2) , 1);
                pf_frequency_noise_default[view]->Fill((j + 0.5) * fFFTBinWidth*2, h1_fft->GetBinContent(2*j+1) + h1_fft->GetBinContent(2*j+2), 1);
              }
            }

            // chi-square according to template
            if (!fRebin) {
              for (int j = 0; j < fWindowSize/2; ++j) {
                double value_noise = (h1_fft->GetBinContent(j+1) - pf_frequency_noise_template[view]->GetBinContent(j+1)) / (pf_frequency_noise_template[view]->GetBinError(j+1));
                noisefrequency_noise[channel][k][j] = value_noise*value_noise;

                double value_signal = (pf_frequency_signal_template[view]->GetBinContent(j+1)) / (pf_frequency_noise_template[view]->GetBinError(j+1));
                noisefrequency_signal[channel][k][j] = value_signal*value_signal;

                double value_cross = value_noise * value_signal;
                noisefrequency_cross[channel][k][j] = value_cross;
              }
            }
            else {
              for (int j = 0; j < fWindowSize/4; ++j) {
                double value_noise = (h1_fft->GetBinContent(2*j+1) + h1_fft->GetBinContent(2*j+2) - pf_frequency_noise_template[view]->GetBinContent(j+1)) / (pf_frequency_noise_template[view]->GetBinError(j+1));
                noisefrequency_noise[channel][k][j] = value_noise*value_noise;

                double value_signal = (pf_frequency_signal_template[view]->GetBinContent(j+1)) / (pf_frequency_noise_template[view]->GetBinError(j+1));
                noisefrequency_signal[channel][k][j] = value_signal*value_signal;

                double value_cross = value_noise * value_signal;
                noisefrequency_cross[channel][k][j] = value_cross;
              }
            }

            delete h1_waveform;
            delete h1_fft;
          } // fDoFrequencyTest

          if (fDoCNNROI) {
            for (size_t ii=0; ii<cnnroiwirelist.size(); ++ii) {
              if (int(cnnroiwirelist[ii]->Channel()) != channel) continue;
              std::vector<float> vecScore = cnnroiwirelist[ii]->Signal();

              double score = vecScore[start_tick+1]; // each window have same score
              if (score >= 1) score = 1. - 1e-7;
              h1_noisecnnscore[view]->Fill(score); 
            } // loop over cnnroiwirelist ii
          } // fDoCNNROI

        } // loop over k
      } // if not fUseFullWaveform
    } // loop over rawgigit or wire
  } // fSaveNoise
  
  // ---------- evaluate a waveform ----------
  if (fEvaluateWaveform) {
    for (size_t i = 0; i < (rawdigitlist.empty() ? wirelist.size() : rawdigitlist.size()); ++i) {

      // test waveform
      std::vector<float> adcvec(fNTicks);
      int channel = -1;

      if (rawdigitlist.size()) {
        channel = rawdigitlist[i]->Channel();
        if (chStatus.IsBad(channel)) continue;

        std::vector<short> rawadc(fNTicks);
        raw::Uncompress(rawdigitlist[i]->ADCs(), rawadc, rawdigitlist[i]->GetPedestal(), rawdigitlist[i]->Compression());
        for (size_t j = 0; j < rawadc.size(); ++j) {
          adcvec[j] = rawadc[j] - rawdigitlist[i]->GetPedestal();
          if (fSaveWaveform && fSaveTree) channeladc[channel][j] = adcvec[j];
        }
      }
      else if (wirelist.size()) {
        channel = wirelist[i]->Channel();
        if (chStatus.IsBad(channel)) continue;

        const auto& signal = wirelist[i]->Signal();
        if (fUseFloat) {
          for (size_t j = 0; j < adcvec.size(); ++j) {
            adcvec[j] = signal[j];
            if (fSaveWaveform && fSaveTree) channeladc[channel][j] = adcvec[j];
          }
        }
        else {
          for (size_t j = 0; j < adcvec.size(); ++j) {
            adcvec[j] = short(signal[j]);
            if (fSaveWaveform && fSaveTree) channeladc[channel][j] = adcvec[j];
          }
        }
      }

      int view = geo->View(channel);

      bool IsSignalWire = false;
      std::vector<int> signal_starttick;
      std::vector<int> signal_endtick;
      std::vector<double> signal_max; // according to number of electrons or energy deposits
      std::vector<int> signal_max_tdctick;
      for (auto const& simchannel : (*simChannelHandle)) {
        // .. get simChannel channel number
        if (int(simchannel.Channel()) != channel) continue;
        
        getSignal(simchannel, signal_starttick, signal_endtick, signal_max, signal_max_tdctick);

        if (signal_starttick.size() == 0 ||
            (signal_endtick.size() == 1 && signal_endtick.back() == -999))
          continue;

        IsSignalWire = true;
        
        if (fSaveTree) {
          simchannelvec.push_back(channel);
          signalstarttickvec.push_back(signal_starttick);
          signalendtickvec.push_back(signal_endtick);
          signalmaxtickvec.push_back(signal_max_tdctick);
          signalmaxvec.push_back(signal_max);
        }
      } // loop over simchannel


      if (fUseFullWaveform) {
        cout << "www:  EvaluateWaveform using full waveform needs to be added!" << endl;
      } // if fUseFullWaveform
      else {
        int nStrides = int(fNTicks / fStrideSize + 1);
        if (fDropLastWindow) nStrides = nStrides - 1;

        for (int k = 0; k < nStrides; ++k) {
          int start_tick = fStrideSize * k;
          int end_tick = start_tick + fWindowSize - 1;
          if (end_tick > fNTicks - 1) {
            end_tick = fNTicks - 1;
            start_tick = end_tick + 1 - fWindowSize;
          }
          
          // adc sum
          double adcsum = 0.;
          for (int j = start_tick; j <= end_tick; ++j) {
            adcsum += adcvec[j];
          }
          waveformadcsum[channel][k] = adcsum;
            
          // charge
          float chargesum = 0.;
          double signal = 0.;
          if (IsSignalWire) {
            // signal
            for (size_t s = 0; s < signal_starttick.size(); ++s) {
              if(isSignalInWindow(signal_max_tdctick[s], start_tick, end_tick)) {
                if (signal_max[s] > signal) {
                  signal = signal_max[s];
                }
              }
            }
            // charge sum
            for (auto const& simchannel : (*simChannelHandle)) {
              // .. get simChannel channel number
              if (int(simchannel.Channel()) != channel) continue;
              chargesum = getCharge(simchannel, start_tick, end_tick);
            }
          }
          waveformchargesum[channel][k] = chargesum;
          
          if (fDoADCThreshold) {
            float maxadc =0;
            for (int j=start_tick; j<= end_tick; ++j) {
              if (adcvec[j] > maxadc) maxadc = adcvec[j];
            }
            if (fCheckWaveform) waveformmaxadc[channel][k] = maxadc;

            // signal
            if (signal > 0) {
              h1_signalmaxadc_all[view]->Fill(signal);
              if (maxadc > fAdcThresholdCut[view]) {
                h1_signalmaxadc_select[view]->Fill(signal);
              }

              h1_maxadc_signal[view]->Fill(maxadc);
            }
            else {
              h1_maxadc_noise[view]->Fill(maxadc);
            }

            // charge sum
            if (chargesum > 0) {
              h1_esummaxadc_all[view]->Fill(chargesum);
              if (maxadc > fAdcThresholdCut[view]) {
                h1_esummaxadc_select[view]->Fill(chargesum);
              }

              h1_esummaxadc_signal[view]->Fill(maxadc);
            }
            else {
              h1_esummaxadc_noise[view]->Fill(maxadc);
            }

            if (!IsSignalWire) {
              h1_maxadc_noise_pure[view]->Fill(maxadc);
              h1_esummaxadc_noise_pure[view]->Fill(maxadc);
            }
          } // fDoADCThreshold

          if (fDoCNNROI) {
            for (size_t ii=0; ii<cnnroiwirelist.size(); ++ii) {
              if (int(cnnroiwirelist[ii]->Channel()) != channel) continue;
              std::vector<float> vecScore = cnnroiwirelist[ii]->Signal();

              double score = vecScore[start_tick+1]; // each window have same score

              if (fCheckWaveform) waveformcnnscore[channel][k] = score;

              // signal
              if (signal > 0) {
                h1_signalcnnroi_all[view]->Fill(signal);
                if (score > fROIScoreCut[view]) {
                  h1_signalcnnroi_select[view]->Fill(signal);
                }

                h1_roiscore_signal[view]->Fill(score);
              }
              else {
                h1_roiscore_noise[view]->Fill(score);
              }

              // charge sum
              if (chargesum > 0) {
                h1_esumcnnroi_all[view]->Fill(chargesum);
                if (score > fROIScoreCut[view]) {
                  h1_esumcnnroi_select[view]->Fill(chargesum);
                }

                h1_esumroiscore_signal[view]->Fill(score);
              }
              else {
                h1_esumroiscore_noise[view]->Fill(score);
              }

              if (!IsSignalWire) {
                h1_roiscore_noise_pure[view]->Fill(score);
                h1_esumroiscore_noise_pure[view]->Fill(score);
              }
            } // loop over cnnroiwirelist ii
          } // fDoCNNROI

          if (fDoFrequencyTest) {
            waveformsignal[channel][k] = signal;

            TH1D* h1_waveform = new TH1D(Form("waveform_%d_%d", channel, k), Form("waveform_%d_%d", channel, k), fWindowSize, 0, fWindowSize);
            for (int j = start_tick; j <= end_tick; ++j) {
              h1_waveform->SetBinContent(j-start_tick+1, adcvec[j]);
            }

            std::vector<float> tdcchargevec(fWindowSize, 0);

            if (IsSignalWire) {
              for (auto const& simchannel : (*simChannelHandle)) {
                // .. get simChannel channel number
                if (int(simchannel.Channel()) != channel) continue;
                if (fCheckWaveform) getTdcCharge(simchannel, start_tick, end_tick, tdcchargevec);
              }
            }

            if (fCheckWaveform) {
              for (int j=0; j<fWindowSize; ++j) {
                waveformtdccharge[channel][k][j] = tdcchargevec[j];
              }
            }

            TH1D* h1_fft = new TH1D(Form("fft_%d_%d", channel, k), Form("fft_%d_%d", channel, k), fWindowSize, 0, fWindowSize*fFFTBinWidth);

            calculateFFT(h1_waveform, h1_fft);

            // chi-square according to template
            for (int j = 0; j < fWindowSize/2; ++j) {
              double value_noise = (h1_fft->GetBinContent(j+1) - pf_frequency_noise_template[view]->GetBinContent(j+1)) / (pf_frequency_noise_template[view]->GetBinError(j+1));
              frequency_noise[channel][k][j] = value_noise*value_noise;

              double value_signal = (pf_frequency_signal_template[view]->GetBinContent(j+1)) / (pf_frequency_noise_template[view]->GetBinError(j+1));
              frequency_signal[channel][k][j] = value_signal*value_signal;

              double value_cross = value_noise * value_signal;
              frequency_cross[channel][k][j] = value_cross;
            }

            delete h1_waveform;
            delete h1_fft;
          }
        } // loop over nStrides k
      } // use short waveform
    } // loop over rawdigitlist or wirelist  

  } // if fEvaluateWaveform
  
  if (fSaveTree) fTree->Fill();
}

void SignalNoiseAna::beginJob()
{
  art::ServiceHandle<art::TFileService const> tfs;

  // frequencytest template
  TFile TemplateFFTFile(fFrequencyTemplateFile.c_str());
  for (int i=0; i<fNPlanes; i++) {
    if (fUseFloat) {
      pf_frequency_noise_template[i] = (TProfile*)TemplateFFTFile.Get(Form("pf_frequency_noise_float_%d",i));
      pf_frequency_signal_template[i] = (TProfile*)TemplateFFTFile.Get(Form("pf_frequency_signal_float_%d",i));
    }
    else {
      pf_frequency_noise_template[i] = (TProfile*)TemplateFFTFile.Get(Form("pf_frequency_noise_short_%d",i));
      pf_frequency_signal_template[i] = (TProfile*)TemplateFFTFile.Get(Form("pf_frequency_signal_short_%d",i));
    }
  }

  // ttree
  if (fSaveTree) {
    fTree = tfs->make<TTree>("anatree", "signalnoiseana");
    fTree->Branch("run", &run, "run/I");
    fTree->Branch("subrun", &subrun, "subrun/I");
    fTree->Branch("event", &event, "event/I");

    if (fSaveWaveform) fTree->Branch("channeladc", channeladc, "channeladc[480][2048]/F");

    if (fSaveNoise) {

      if (fDoFrequencyTest) {
        fTree->Branch("noisefrequency_noise", noisefrequency_noise, "noisefrequency_noise[480][10][100]/F");
        fTree->Branch("noisefrequency_signal", noisefrequency_signal, "noisefrequency_signal[480][10][100]/F");
        fTree->Branch("noisefrequency_cross", noisefrequency_cross, "noisefrequency_cross[480][10][100]/F");

      }
    } // fSaveNoise
  } // fSaveTree

  for (int i = 0; i < fNPlanes; ++i) {
    if (fUseNumElectrons) {
      h1_signal[i] = tfs->make<TH1D>(Form("h1_signal_%d", i), Form("Plane %d; Number of Electrons;", i), 100, 0, 20000);
      h2_signalADC_sametick[i] = tfs->make<TH2D>(Form("h2_signalADC_sametick_%d",i), Form("Plane %d; number of electrons; ADC counts", i), 100, 0, 20000, 50, 0, 50);
      h2_signalADC_max[i] = tfs->make<TH2D>(Form("h2_signalADC_max_%d",i), Form("Plane %d; number of electrons; ADC counts", i), 100, 0, 20000, 50, 0, 50);
    }
    else {
      h1_signal[i] = tfs->make<TH1D>(Form("h1_signal_%d", i), Form("Plane %d; Energy [MeV];", i), 100, 0, 1);
      h2_signalADC_sametick[i] = tfs->make<TH2D>(Form("h2_signalADC_sametick_%d",i), Form("Plane %d; Energy [MeV]; ADC counts", i), 100, 0, 1, 50, 0, 50);
      h2_signalADC_max[i] = tfs->make<TH2D>(Form("h2_signalADC_max_%d",i), Form("Plane %d; Energy [MeV]; ADC counts", i), 100, 0, 1, 50, 0, 50);
    }
  
    h1_tickdiff_max[i] = tfs->make<TH1D>(Form("h1_tickdiff_max_%d", i), Form("Plane %d; tick diff (maxE - max pulse);", i), 100, -50, 50);
  
    if (fSaveNoise) {
      h1_noiserms[i] = tfs->make<TH1D>(Form("h1_noiserms_%d",i), Form("Plane %d; ADC counts;", i), 200, 0, 10);
      h1_noisemaxadc[i] = tfs->make<TH1D>(Form("h1_noisemaxadc_%d", i), Form("Plane %d; Max adc; Count [#]", i), 200, 0, 20);
      h1_noisecnnscore[i] = tfs->make<TH1D>(Form("h1_noisecnnscore_%d", i), Form("Plane %d; cnn score; Count [#]", i), 100000, 0, 1);
    }

    if (fUseFullWaveform) {
      // TProfile: option "s" uses RMS as error; defauls uses RMS/sqrt(N) as error
      pf_frequency_noise[i] = tfs->make<TProfile>(Form("pf_frequency_noise_%d",i), Form("Plane %d; Frequency [kHz]; Magnitude", i), fNTicks, 0, fNTicks*fFFTBinWidth, "s");
      pf_frequency_signal[i] = tfs->make<TProfile>(Form("pf_frequency_signal_%d",i), Form("Plane %d; Frequency [kHz]; Magnitude", i), fNTicks, 0, fNTicks*fFFTBinWidth, "s");

      pf_frequency_noise_default[i] = tfs->make<TProfile>(Form("pf_frequency_noise_default_%d",i), Form("Plane %d; Frequency [kHz]; Magnitude", i), fNTicks, 0, fNTicks*fFFTBinWidth);
      pf_frequency_signal_default[i] = tfs->make<TProfile>(Form("pf_frequency_signal_default_%d",i), Form("Plane %d; Frequency [kHz]; Magnitude", i), fNTicks, 0, fNTicks*fFFTBinWidth);
    }
    else {
      if (!fRebin) {
        pf_frequency_noise[i] = tfs->make<TProfile>(Form("pf_frequency_noise_%d",i), Form("Plane %d; Frequency [kHz]; Magnitude", i), fWindowSize, 0, fWindowSize*fFFTBinWidth, "s");
        pf_frequency_signal[i] = tfs->make<TProfile>(Form("pf_frequency_signal_%d",i), Form("Plane %d; Frequency [kHz]; Magnitude", i), fWindowSize, 0, fWindowSize*fFFTBinWidth, "s");

        pf_frequency_noise_default[i] = tfs->make<TProfile>(Form("pf_frequency_noise_default_%d",i), Form("Plane %d; Frequency [kHz]; Magnitude", i), fWindowSize, 0, fWindowSize*fFFTBinWidth);
        pf_frequency_signal_default[i] = tfs->make<TProfile>(Form("pf_frequency_signal_default_%d",i), Form("Plane %d; Frequency [kHz]; Magnitude", i), fWindowSize, 0, fWindowSize*fFFTBinWidth);
      }
      else {
        pf_frequency_noise[i] = tfs->make<TProfile>(Form("pf_frequency_noise_%d",i), Form("Plane %d; Frequency [kHz]; Magnitude", i), fWindowSize/2, 0, fWindowSize*fFFTBinWidth, "s");
        pf_frequency_signal[i] = tfs->make<TProfile>(Form("pf_frequency_signal_%d",i), Form("Plane %d; Frequency [kHz]; Magnitude", i), fWindowSize/2, 0, fWindowSize*fFFTBinWidth, "s");

        pf_frequency_noise_default[i] = tfs->make<TProfile>(Form("pf_frequency_noise_default_%d",i), Form("Plane %d; Frequency [kHz]; Magnitude", i), fWindowSize, 0, fWindowSize/2*fFFTBinWidth);
        pf_frequency_signal_default[i] = tfs->make<TProfile>(Form("pf_frequency_signal_default_%d",i), Form("Plane %d; Frequency [kHz]; Magnitude", i), fWindowSize, 0, fWindowSize/2*fFFTBinWidth);
      }
    }

    if (fEvaluateWaveform) {

      h1_signalmaxadc_all[i] = tfs->make<TH1D>(Form("h1_signalmaxadc_all_%d", i), Form("Plane %d; Number of electrons/tick/wire; Count [#]", i), 100, 0, 20000);
      h1_signalmaxadc_select[i] = tfs->make<TH1D>(Form("h1_signalmaxadc_select_%d", i), Form("Plane %d; Number of electrons/tick/wire; Count [#]", i), 100, 0, 20000);
      h1_maxadc_signal[i] = tfs->make<TH1D>(Form("h1_maxadc_signal_%d", i), Form("Plane %d; Max adc; Count [#]", i), 2000, 0, 200);
      h1_maxadc_noise[i] = tfs->make<TH1D>(Form("h1_maxadc_noise_%d", i), Form("Plane %d; Max adc; Count [#]", i), 500, 0, 50);
      h1_maxadc_noise_pure[i] = tfs->make<TH1D>(Form("h1_maxadc_noise_pure_%d", i), Form("Plane %d; Max adc; Count [#]", i), 200, 0, 20);

      h1_signalcnnroi_all[i] = tfs->make<TH1D>(Form("h1_signalcnnroi_all_%d", i), Form("Plane %d; Number of electrons/tick/wire; Count [#]", i), 100, 0, 20000);
      h1_signalcnnroi_select[i] = tfs->make<TH1D>(Form("h1_signalcnnroi_select_%d", i), Form("Plane %d; Number of electrons/tick/wire; Count [#]", i), 100, 0, 20000);
      h1_roiscore_signal[i] = tfs->make<TH1D>(Form("h1_roiscore_signal_%d", i), Form("Plane %d; roi score; Count [#]", i), 10000, 0, 1);
      h1_roiscore_noise[i] = tfs->make<TH1D>(Form("h1_roiscore_noise_%d", i), Form("Plane %d; roi score; Count [#]", i), 10000, 0, 1);
      h1_roiscore_noise_pure[i] = tfs->make<TH1D>(Form("h1_roiscore_noise_pure_%d", i), Form("Plane %d; roi score; Count [#]", i), 10000, 0, 1);

      // sum of electrons
      h1_esummaxadc_all[i] = tfs->make<TH1D>(Form("h1_esummaxadc_all_%d", i), Form("Plane %d; Number of electrons; Count [#]", i), 100, 0, 100000);
      h1_esummaxadc_select[i] = tfs->make<TH1D>(Form("h1_esummaxadc_select_%d", i), Form("Plane %d; Number of electrons; Count [#]", i), 100, 0, 100000);
      h1_esummaxadc_signal[i] = tfs->make<TH1D>(Form("h1_esummaxadc_signal_%d", i), Form("Plane %d; Max adc; Count [#]", i), 2000, 0, 200);
      h1_esummaxadc_noise[i] = tfs->make<TH1D>(Form("h1_esummaxadc_noise_%d", i), Form("Plane %d; Max adc; Count [#]", i), 2000, 0, 200);
      h1_esummaxadc_noise_pure[i] = tfs->make<TH1D>(Form("h1_esummaxadc_noise_pure_%d", i), Form("Plane %d; Max adc; Count [#]", i), 2000, 0, 200);

      h1_esumcnnroi_all[i] = tfs->make<TH1D>(Form("h1_esumcnnroi_all_%d", i), Form("Plane %d; Number of electrons; Count [#]", i), 100, 0, 100000);
      h1_esumcnnroi_select[i] = tfs->make<TH1D>(Form("h1_esumcnnroi_select_%d", i), Form("Plane %d; Number of electrons; Count [#]", i), 100, 0, 100000);
      h1_esumroiscore_signal[i] = tfs->make<TH1D>(Form("h1_esumroiscore_signal_%d", i), Form("Plane %d; roi score; Count [#]", i), 10000, 0, 1);
      h1_esumroiscore_noise[i] = tfs->make<TH1D>(Form("h1_esumroiscore_noise_%d", i), Form("Plane %d; roi score; Count [#]", i), 10000, 0, 1);
      h1_esumroiscore_noise_pure[i] = tfs->make<TH1D>(Form("h1_esumroiscore_noise_pure_%d", i), Form("Plane %d; roi score; Count [#]", i), 10000, 0, 1);
    }
  }

  if (fEvaluateWaveform) {
    fTree->Branch("waveformadcsum", waveformadcsum, "waveformadcsum[480][10]/F");
    fTree->Branch("waveformchargesum", waveformchargesum, "waveformchargesum[480][10]/F");
    fTree->Branch("waveformsignal", waveformsignal, "waveformsignal[480][10]/F");
    fTree->Branch("frequency_noise", frequency_noise, "frequency_noise[480][10][100]/F");
    fTree->Branch("frequency_signal", frequency_signal, "frequency_signal[480][10][100]/F");
    fTree->Branch("frequency_cross", frequency_cross, "frequency_cross[480][10][100]/F");

    if (fCheckWaveform) {
      fTree->Branch("waveformmaxadc", waveformmaxadc, "waveformmaxadc[480][10]/F");
      fTree->Branch("waveformcnnscore", waveformcnnscore, "waveformcnnscore[480][10]/F");
      fTree->Branch("waveformtdccharge", waveformtdccharge, "waveformtdccharge[480][10][200]/F");
    }

    fTree->Branch("simchannelvec", &simchannelvec);
    fTree->Branch("signalstarttickvec", &signalstarttickvec);
    fTree->Branch("signalendtickvec", &signalendtickvec);
    fTree->Branch("signalmaxtickvec", &signalmaxtickvec);
    fTree->Branch("signalmaxvec", &signalmaxvec);
  }

}


void SignalNoiseAna::endJob()
{
  if (fEvaluateWaveform) {
    art::ServiceHandle<art::TFileService const> tfs;

    for (int p = 0; p < fNPlanes; ++p) {
      if (TEfficiency::CheckConsistency(*(h1_signalmaxadc_select[p]), *(h1_signalmaxadc_all[p]))) {
        eff_maxadc[p] = tfs->make<TEfficiency>(*(h1_signalmaxadc_select[p]), *(h1_signalmaxadc_all[p]));
        eff_maxadc[p]->Write(Form("eff_maxadc_%d", p));
      }

      if (TEfficiency::CheckConsistency(*(h1_signalcnnroi_select[p]), *(h1_signalcnnroi_all[p]))) {
        eff_cnnroi[p] = tfs->make<TEfficiency>(*(h1_signalcnnroi_select[p]), *(h1_signalcnnroi_all[p]));
        eff_cnnroi[p]->Write(Form("eff_cnnroi_%d", p));
      }

      if (TEfficiency::CheckConsistency(*(h1_esummaxadc_select[p]), *(h1_esummaxadc_all[p]))) {
        eff_esummaxadc[p] = tfs->make<TEfficiency>(*(h1_esummaxadc_select[p]), *(h1_esummaxadc_all[p]));
        eff_esummaxadc[p]->Write(Form("eff_esummaxadc_%d", p));
      }

      if (TEfficiency::CheckConsistency(*(h1_esumcnnroi_select[p]), *(h1_esumcnnroi_all[p]))) {
        eff_esumcnnroi[p] = tfs->make<TEfficiency>(*(h1_esumcnnroi_select[p]), *(h1_esumcnnroi_all[p]));
        eff_esumcnnroi[p]->Write(Form("eff_esumcnnroi_%d", p));
      }

    }
  } 
}


void SignalNoiseAna::calculateFFT(TH1D* h1_waveform, TH1D* h1_fft) {
  int nbins = h1_waveform->GetNbinsX();

  TH1* h1_transform = 0;

  h1_transform = h1_waveform->FFT(h1_transform, "MAG");

  // loop over h1_transform
  for (int k=0; k<nbins; k++) {
    double frequency = (k+0.5) * fFFTBinWidth; // [kHz]
    double amplitude = h1_transform->GetBinContent(k+1);
    h1_fft->Fill(frequency, amplitude);
  }

  h1_transform->Delete();
}


void SignalNoiseAna::getSignal(const sim::SimChannel& simchannel, std::vector<int> & signal_starttick, std::vector<int> & signal_endtick, std::vector<double> & signal_max, std::vector<int> & signal_max_tdctick) {

  auto const clockData = art::ServiceHandle<detinfo::DetectorClocksService>()->DataForJob();
  auto const detProp = art::ServiceHandle<detinfo::DetectorPropertiesService>()->DataForJob(clockData);

  auto const& timeSlices = simchannel.TDCIDEMap();

  // time slice from simChannel is for individual tick
  // group neighboring time slices into a "signal"
  int tdctick_previous = -999;
  double maxE = -999.; // number of electrons or energy deposits
  int maxEtick = -999;
  
  for (auto const& timeSlice : timeSlices) {
    auto const tpctime = timeSlice.first;
    auto const& energyDeposits = timeSlice.second;
    int tdctick = static_cast<int>(clockData.TPCTDC2Tick(double(tpctime)));
    if (tdctick < 0 || tdctick > int(detProp.ReadOutWindowSize()) - 1) continue;

    // for a time slice, there may exist more than one energy deposit.
    double tdc_signal = 0;
    for (auto const& energyDeposit : energyDeposits) {
      if (fUseNumElectrons) {
        tdc_signal += energyDeposit.numElectrons;
      }
      else {
        tdc_signal += energyDeposit.energy;
      }
    }

    if (tdctick_previous == -999) {
      signal_starttick.push_back(tdctick);
      maxE = tdc_signal;
      maxEtick = tdctick;
    }
    else if (tdctick - tdctick_previous != 1) {
      signal_starttick.push_back(tdctick);
      signal_endtick.push_back(tdctick_previous);
      signal_max.push_back(maxE);
      signal_max_tdctick.push_back(maxEtick);
      maxE = tdc_signal;
      maxEtick = tdctick;
    }
    else if (tdctick - tdctick_previous == 1) {
      if (maxE < tdc_signal) {
        maxE = tdc_signal;
        maxEtick = tdctick;
      }
    }

    tdctick_previous = tdctick;

  } // loop over timeSlices timeSlice

  signal_endtick.push_back(tdctick_previous); // for last one
  signal_max.push_back(maxE);          // for last one
  signal_max_tdctick.push_back(maxEtick);     // for last one
}


void SignalNoiseAna::getTdcCharge(const sim::SimChannel& simchannel, int starttick, int endtick, std::vector<float> & tdcchargevec) {

  auto const clockData = art::ServiceHandle<detinfo::DetectorClocksService>()->DataForJob();
  auto const detProp = art::ServiceHandle<detinfo::DetectorPropertiesService>()->DataForJob(clockData);

  auto const& timeSlices = simchannel.TDCIDEMap();

  for (auto const& timeSlice : timeSlices) {
    auto const tpctime = timeSlice.first;
    auto const& energyDeposits = timeSlice.second;
    int tdctick = static_cast<int>(clockData.TPCTDC2Tick(double(tpctime)));
    if (tdctick < starttick || tdctick > endtick) continue;

    // for a time slice, there may exist more than one energy deposit.
    double charge = 0;
    for (auto const& energyDeposit : energyDeposits) {
      if (fUseNumElectrons) {
        charge += energyDeposit.numElectrons;
      }
      else {
        charge += energyDeposit.energy;
      }
    }
    tdcchargevec[tdctick-starttick] = charge;
  }
}


float SignalNoiseAna::getCharge(const sim::SimChannel& simchannel, int starttick, int endtick) {

  auto const clockData = art::ServiceHandle<detinfo::DetectorClocksService>()->DataForJob();
  auto const detProp = art::ServiceHandle<detinfo::DetectorPropertiesService>()->DataForJob(clockData);

  auto const& timeSlices = simchannel.TDCIDEMap();

  float charge = 0.;
  for (auto const& timeSlice : timeSlices) {
    auto const tpctime = timeSlice.first;
    auto const& energyDeposits = timeSlice.second;
    int tdctick = static_cast<int>(clockData.TPCTDC2Tick(double(tpctime)));
    if (tdctick < starttick || tdctick > endtick) continue;

    // for a time slice, there may exist more than one energy deposit.
    for (auto const& energyDeposit : energyDeposits) {
      if (fUseNumElectrons) {
        charge += energyDeposit.numElectrons;
      }
      else {
        charge += energyDeposit.energy;
      }
    }
  }
  return charge;
}


bool SignalNoiseAna::isSignalInWindow(int maxtick, int starttick, int endtick) {
  // here, endtick included in the range
  return starttick <= maxtick && maxtick <= endtick;
}


void SignalNoiseAna::reset() {
  run = -99;
  subrun = -99;
  event = -99;
  
  for (int i = 0; i < 480; ++i) {
    for (int j = 0; j < 2048; ++j) {
      channeladc[i][j] = -9999;
    }

    for (int k = 0; k < 10; ++k) {
      if (fSaveNoise) {
        for (int n=0; n<100; n++) {
          noisefrequency_noise[i][k][n] = -9999;
          noisefrequency_signal[i][k][n] = -9999;
          noisefrequency_cross[i][k][n] = -9999;
        }
      }

      waveformadcsum[i][k] = -9999;
      waveformchargesum[i][k] = -9999;
      waveformsignal[i][k] = -9999;
      
      for (int n=0; n<100; n++) {
        frequency_noise[i][k][n] = -9999;
        frequency_signal[i][k][n] = -9999;
        frequency_cross[i][k][n] = -9999;
      }

      if (fCheckWaveform) {
        waveformmaxadc[i][k] = -9999;
        waveformcnnscore[i][k] = -9999;
        for (int n=0; n<100; n++) {
          waveformtdccharge[i][k][n] = -9999;
        }
      }
    }

  }

  simchannelvec.clear();
  signalstarttickvec.clear();
  signalendtickvec.clear();
  signalmaxtickvec.clear();
  signalmaxvec.clear();
}

DEFINE_ART_MODULE(SignalNoiseAna)
